package de.hft.stuttgart.citydoctor2.healer;

import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.ade.ADEException;
import org.citygml4j.core.model.CityGMLVersion;
import org.citygml4j.core.model.core.AbstractCityObject;
import org.citygml4j.core.model.core.AbstractCityObjectProperty;
import org.citygml4j.core.model.core.CityModel;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.citygml4j.xml.CityGMLContext;
import org.citygml4j.xml.writer.CityGMLOutputFactory;
import org.citygml4j.xml.writer.CityGMLWriteException;
import org.citygml4j.xml.writer.CityGMLWriter;

import de.hft.stuttgart.citydoctor2.CityDoctorValidation;
import de.hft.stuttgart.citydoctor2.check.AbstractCheck;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checker;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.ValidationConfiguration;
import de.hft.stuttgart.citydoctor2.check.error.AttributeMissingError;
import de.hft.stuttgart.citydoctor2.check.error.AttributeValueWrongError;
import de.hft.stuttgart.citydoctor2.checkresult.utility.CheckReportWriteException;
import de.hft.stuttgart.citydoctor2.checks.SvrlContentHandler;
import de.hft.stuttgart.citydoctor2.datastructure.CityDoctorModel;
import de.hft.stuttgart.citydoctor2.datastructure.CityObject;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.exceptions.CityDoctorWriteException;
import de.hft.stuttgart.citydoctor2.parameter.ArgumentParser;
import de.hft.stuttgart.citydoctor2.parser.CityGmlConsumer;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParseException;
import de.hft.stuttgart.citydoctor2.parser.CityGmlParser;
import de.hft.stuttgart.citydoctor2.parser.InvalidGmlFileException;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.reporting.XmlStreamReporter;
import de.hft.stuttgart.citydoctor2.reporting.pdf.PdfStreamReporter;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import de.hft.stuttgart.quality.QualityADEModule;

public class Healer {

	private static final Logger logger = LogManager.getLogger(Healer.class);

	private int numIterations = 200;
	private int schematronIterations = 2;

	private Checker checker;
	private ModificationListener l;
	private HealingPlan plan;

	static {
//		CppInitializer.initCpp();
	}

	public static void main(String[] args) throws CityGmlParseException, IOException,
			InvalidGmlFileException, CityDoctorWriteException {
		ArgumentParser argParser = new ArgumentParser(args);
		String inputFile = CityDoctorValidation.getInputFile(argParser, false);
		String outputFile = getOutputFile(argParser);
		String xmlOutput = CityDoctorValidation.getXmlOutput(argParser);
		String pdfOutput = CityDoctorValidation.getPdfOutput(argParser);
		ValidationConfiguration config = CityDoctorValidation.getValidationConfig(argParser, true);
		startHealingProcess(inputFile, outputFile, xmlOutput, pdfOutput, config);
	}

	private static void startHealingProcess(String inputFile, String outputFile, String xmlOutput, String pdfOutput,
			ValidationConfiguration config) throws CityGmlParseException, IOException,
			InvalidGmlFileException, CityDoctorWriteException {
		ParserConfiguration parserConfig = config.getParserConfiguration();

		if (config.isUseStreaming()) {
			Healer.streamHeal(new File(inputFile), outputFile, xmlOutput, pdfOutput, config);
		} else {
			CityDoctorModel model = CityGmlParser.parseCityGmlFile(inputFile, parserConfig);
			Checker c = new Checker(config, model);
			Healer healer = new Healer(c, (ModificationListener) null);
			healer.healModel(model);
			model.saveAs(outputFile, true);
		}
	}

	public static void streamHeal(File inputFile, String outputFileName, String xmlOutput, String pdfOutput,
			ValidationConfiguration config) throws CityGmlParseException, IOException {
		if (!inputFile.exists() || !inputFile.isFile()) {
			logger.error("Inputfile does not exist or is not a file: {}", inputFile.getAbsolutePath());
		}
		Checker c = new Checker(config, null);
		Healer healer = new Healer(c, (ModificationListener) null);
		try (BufferedOutputStream xmlBos = Checker.getXmlOutputMaybe(xmlOutput);
				BufferedOutputStream pdfBos = Checker.getPdfOutputMaybe(pdfOutput)) {
			String fileName = inputFile.getName();
			XmlStreamReporter xmlReporter;
			if (xmlBos != null) {
				xmlReporter = new XmlStreamReporter(xmlBos, fileName, config);
			} else {
				xmlReporter = null;
			}
			PdfStreamReporter pdfReporter;
			if (pdfBos != null) {
				pdfReporter = new PdfStreamReporter(pdfBos, fileName, config);
			} else {
				pdfReporter = null;
			}

			CityGmlConsumer coConsumer = new CityGmlConsumer() {

				@Override
				public void accept(CityObject co) {
					healer.healCityObject(co);
				}

			};
			CityGmlParser.streamCityGml(inputFile, config.getParserConfiguration(), coConsumer, outputFileName);
			Checker.writeReport(xmlReporter);
			Checker.writeReport(pdfReporter);
		} catch (CheckReportWriteException e) {
			logger.error(Localization.getText("Checker.failReports"), e);
		}
	}

	private static String getOutputFile(ArgumentParser argParser) {
		String outputFile;
		if (!argParser.containsOption("out")) {
			logger.error("No output file specified. (-out [FILE])");
			System.exit(10);
		}
		List<String> outFiles = argParser.getValues("out");
		if (outFiles.size() != 1) {
			logger.error("Specify exactly one file as output.");
			System.exit(11);
		}
		outputFile = outFiles.get(0);
		return outputFile;
	}

	public static void heal(File in, File configFile, File out)
			throws FileNotFoundException, CityGmlParseException, InvalidGmlFileException, CityDoctorWriteException {
		ValidationConfiguration config;
		if (configFile == null) {
			config = ValidationConfiguration.loadStandardValidationConfig();
		} else {
			config = ValidationConfiguration.loadValidationConfig(configFile.getAbsolutePath());
		}
		CityDoctorModel model = CityGmlParser.parseCityGmlFile(in.getAbsolutePath(), config.getParserConfiguration());
		Checker c = new Checker(config, model);
		Healer healer = new Healer(c, (ModificationListener) null);
		healer.healModel(model);
		model.saveAs(out.getAbsolutePath(), true);
	}

	public Healer(Checker c) {
		this(c, (ModificationListener) null);
	}

	public Healer(Checker c, ModificationListener l) {
		plan = HealingMethods.createHealingPlanFromAllMethods();
		checker = c;
		if (l == null) {
			// create useless modification listener to avoid having to check on == null
			l = new ModificationListener() {

				@Override
				public void polygonRemoved(Polygon p) {
					// not observing anything
				}

				@Override
				public void polygonAdded(Polygon p) {
					// not observing anything
				}
			};
		}
		this.l = l;
	}

	public Healer(Checker c, ModificationListener l, HealingPlan plan) {
		this(c, l);
		this.plan = Objects.requireNonNull(plan);
	}

	public Healer(Checker c, HealingPlan plan) {
		this(c);
		this.plan = Objects.requireNonNull(plan);
	}

	public void setNumberOfIterations(int iterations) {
		if (iterations < 1) {
			throw new IllegalArgumentException("Iterations may not be lower than 1");
		}
		numIterations = iterations;
	}

	public void healModel(CityDoctorModel model) {
		model.createFeatureStream().forEach(this::healCityObject);
	}

	public void healCityObject(CityObject co) {
		logger.info("Repairing feature {}", co.getGmlId());
		String schematronPath = checker.getConfig().getSchematronFilePath();
		if (schematronPath != null && !schematronPath.isEmpty()) {
			healSemanticErrors(co);
		}
		// healing geometries
		GeometryFilter filter = new GeometryFilter(g -> heal(g, co));
		co.accept(filter);
	}

	private void healSemanticErrors(CityObject co) {
		// if a schematron file has been declared this needs to execute
		for (int i = 0; i < schematronIterations; i++) {
			co.clearAllContainedCheckResults();
			executeSchematron(co);
			List<CheckError> errors = new ArrayList<>();
			co.collectContainedErrors(errors);
			boolean healedSomething = false;
			for (CheckError err : errors) {
				if (err instanceof AttributeMissingError || err instanceof AttributeValueWrongError) {
					for (HealingMethod m : plan.getHealingMethods()) {
						if (err.accept(m, l)) {
							healedSomething = true;
							break;
						}
					}
				}
			}
			if (!healedSomething) {
				// nothing was changed, don't check for anything more
				break;
			}
		}

		// recheck for geometry errors as they were removed when executing the
		// schematron stuff
		co.prepareForChecking();
		checker.executeChecksForCheckable(co);
		if (checker.getConfig().getParserConfiguration().useLowMemoryConsumption()) {
			co.clearMetaInformation();
		}
	}

	private void executeSchematron(CityObject co) {
		ValidationConfiguration config = checker.getConfig();
		try {
			GeometryFactory factory = GeometryFactory.newInstance();
			// write geometries in citygml4j datastructure
			co.reCreateGeometries(factory, checker.getConfig().getParserConfiguration());
			byte[] gml = writeCityGml(co.getGmlObject());
			// remove them again to save memory
			co.unsetGmlGeometries();
			SvrlContentHandler handler = Checker.executeSchematronValidationIfAvailable(config,
					new ByteArrayInputStream(gml));
			if (handler != null) {
				handler.getFeatureErrors().computeIfPresent(co.getGmlId().getGmlString(), (key, list) -> {
					Checker.handleSchematronErrorsForCityObject(list, co);
					return list;
				});
			}
		} catch (CityGMLWriteException e) {
			throw new IllegalStateException(e);
		}
	}

	/**
	 * Writes one feature to a byte array.
	 * 
	 * @param aco the feature to write
	 * @return the byte array containing the feature
	 * @throws ADEException
	 * @throws CityGMLBuilderException
	 * @throws CityGMLWriteException
	 */
	private byte[] writeCityGml(AbstractCityObject aco) throws CityGMLWriteException {
		CityModel model = new CityModel();
		model.getCityObjectMembers().add(new AbstractCityObjectProperty(aco));
		CityGMLContext context = CityGmlParser.getContext();
		CityGMLOutputFactory outputFactory = context.createCityGMLOutputFactory(CityGMLVersion.v2_0);
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try (CityGMLWriter writer = outputFactory.createCityGMLWriter(out)) {
			writer.withIndent("  ");
			writer.withDefaultPrefixes();
			writer.withPrefix("qual", QualityADEModule.NAMESPACE_URI);
			writer.withSchemaLocation(QualityADEModule.NAMESPACE_URI,
					QualityADEModule.NAMESPACE_URI + "/qualityAde.xsd");
			writer.withDefaultSchemaLocations();
			writer.write(model);
			writer.flush();
			return out.toByteArray();
		}
	}

	public void heal(Geometry geom, CityObject co) {
		heal(geom, co, numIterations);
	}

	public void heal(Geometry geom, CityObject co, int numIterations) {
		if (checker.getConfig().getParserConfiguration().useLowMemoryConsumption()) {
			co.prepareForChecking();
		}
		if (!co.isValidated()) {
			// check it if it has not been checked yet
			checker.executeChecksForCheckable(co);
		}

		try {
			for (int i = 0; i < numIterations; i++) {
				List<CheckError> errors = collectErrorsFromGeometry(geom);
				if (errors.isEmpty()) {
					break;
				} else {
					logger.debug("Found errors: ");
					for (CheckError err : errors) {
						logger.debug(err.getErrorId());
					}
				}
				boolean cannotBeHealed = executeHealingLoop(errors);

				logger.debug("End heal iteration: {}", (i + 1));
				if (cannotBeHealed) {
					logger.debug("Geometry: {} cannot be healed", geom.getGmlId());
					// break iteration loop, nothing more to be done
					break;
				}
				co.prepareForChecking();
				filterOutDuplicateVertices(co);
				// recheck for errors
				checker.executeChecksForCheckable(co);
			}
		} catch (Exception e) {
			logger.debug("Failed to heal geometry", e);
		} finally {
			if (checker.getConfig().getParserConfiguration().useLowMemoryConsumption()) {
				co.clearMetaInformation();
			}
		}
	}

	private void filterOutDuplicateVertices(CityObject co) {
		Map<Vertex, Vertex> map = new HashMap<>();
		co.accept(new AbstractCheck() {
			@Override
			public void check(LinearRing ring) {
				for (int i = 0; i < ring.getVertices().size(); i++) {
					Vertex v = ring.getVertices().get(i);
					Vertex duplicate = map.get(v);
					if (duplicate == null) {
						map.put(v, v);
					} else if (duplicate != v) {
						ring.getVertices().set(i, duplicate);
					}
				}
			}
		});
	}

	private boolean executeHealingLoop(List<CheckError> errors) {
		for (HealingMethod method : plan.getHealingMethods()) {
			for (CheckError err : errors) {
				if (err.accept(method, l)) {
					// geometry got changed, update, recheck, restart
					return false;
				}
			}
		}
		return true;
	}

	private List<CheckError> collectErrorsFromGeometry(Geometry geom) {
		List<CheckError> errors = new ArrayList<>();
		geom.collectContainedErrors(errors);
		return errors;
	}
}
